
; MMC card in SPI mode basic ops. PC interface code for DOS. Phil.Ruston '07
; v3.03 - with debug output file (all read bytes to "debug.bin")
;
; Din is inverted. IE: Normally high and pulled down to zero when MMC out = 1.
; Card reads/writes data on rising edge of clk, bit order: 7 to 0

;----------------------------------------------------------------------------

pport_address equ 0378h

mmc_clock	equ 0	;port 378
mmc_din	equ 1	;port 378
mmc_cs	equ 2	;port 378
mmc_power equ 3	;port 378
pulldown	equ 4	;port 378 - provides a high level for input pulldown

mmc_dout	equ 6	;port 379

;---------------------------------------------------------------------------

.MODEL  SMALL
.STACK  100h

;--------------------------------------------------------------------------

.DATA
Ideal

counter 		dw 0    		; start of data segment
port_shadow	db 0
hexchars 		db "0123456789ABCDEF"
Old8Ofs		dw 0
Old8Seg		dw 0
orig_imr		db 0
time_constant	dd 0
irq_occured	db 0

;--------------------------------------------------------------------------

sector_data	db 512 DUP (0)
filename		db "xxxxxxxx.xxx",0
sector		dw 0h
savelength	dw 0
handle		dw 0
save_length	dw 0
save_data		db 0

debug_buffer	db 16384 DUP (0)
debug_index	dw 0
debug_filename	db "mmcdebug.bin",0

;--------------------------------------------------------------------------

message1		db 10,13,"*********************************************************",10,13,"*         SPI mode MMC PC interface v3.04 + debug       *",10,13,"*           By Phil at www.retroleum.co.uk '07          *",10,13,"* >> Par Port: 0378h - RUN FROM REAL MODE DOS ONLY! <<  *",10,13
		db "*********************************************************",10,13,10,13,"$"
parerrtxt		db "Usage MMC -r:abcd filename (read hex sector abcd from MMC card to 'filename')",10,13
		db "Usage MMC -w:abcd filename (write hex sector abcd from 'filename' to MMC card)",10,13
		db "Usage MMC -i filename (read MMC card CID / CSD to 'filename')",10,13,10,13,"$"
message2		db "Bye..",10,13,"$"

loadokmsg		db "Source sector loaded from disk..",10,13,"$"
saveokmsg		db "Data saved to disk..",10,13,"$"
loaderrmsg	db "Error loading source data from disk",10,13,"$"
saveerrmsg	db "Error saving MMC data to disk..",10,13,"$"
dbsaveokmsg	db "Saved debug file (all bytes which were read from card).",10,13,"$"

mmcreadinfomsg	db "Reading MMC card info (CSD and CID data)..",10,13,"$"
mmcreadmsg	db "Reading data from MMC sector xxxxh..",10,13,"$"
mmcwritemsg	db "Writing data to MMC sector xxxxh..",10,13,"$"


mmc_ok		db "MMC operation completed sucessfully!",10,13,"$"
mmc_initmsg	db "Power cycling and initializing MMC card..",10,13,"$"
mmc_initokmsg	db "MMC initialized to SPI mode OK",10,13,"$"
mmc_error1	db "MMC error! - Initialization failed",10,13,"$"
mmc_error2	db "MMC error! - Cant read card's CSD",10,13,"$"
mmc_error3	db "MMC error! - Cant read card's CID",10,13,"$"
mmc_error4	db "MMC error! - Read card sector failed",10,13,"$"
mmc_error5	db "MMC error! - CRC error",10,13,"$"
mmc_error6	db "MMC error! - Write error",10,13,"$"
mmc_error7	db "MMC error! - Command failed",10,13,"$"
mmc_error8	db "MMC error! - Busy time out after write",10,13,"$"
mmc_error9	db "MMC error! - Timed out waiting for data token",10,13,"$"

;===========================================================================

masm
;---------------------------------------------------------------------------

.CODE
.386p
Ideal


start: 		mov ax,seg counter
        		mov ds,ax
        		cld
        		mov dx,offset message1
        		mov ah,9
        		int 21h
		
		xor ax,ax
		bts ax,pulldown
		bts ax,mmc_cs
		call update_port
		
		call calibrate_timer
							
		mov ax,cs				;get command line options
	        	sub ax,10h
        		mov es,ax
        		xor ax,ax
        		mov cx,40
        		mov si,130               		;look for "-" char in command line
paralp: 		cmp [byte es:si],"-"     		
        		jne fparam	
        		cmp [byte es:si+1],"w"		;followed by "w"?
        		je write_mmc_sector
        		cmp [byte es:si+1],"r"		;followed by "r"?
        		je read_mmc_sector
        		cmp [byte es:si+1],"i"		;followed by "i"?
        		je read_mmc_info
fparam: 		inc si
        		loop paralp
 		mov dx,offset parerrtxt
        		mov ah,9
        		int 021h
		mov ax,4c00h           		;return to dos
        		int 21h

;---------------------------------------------------------------------------------------------------------

		
write_mmc_sector:	add si,3		
		call txt_to_hex
		mov [ds:sector],bx
		add si,5
		call get_filename
		call load_data_from_disk
        		cmp al,1
        		je quit
        		mov dx,[ds:sector]
        		mov di,offset mmcwritemsg+27
        		call hextochars

        		mov dx,offset mmc_initmsg
        		mov ah,9
        		int 21h
        		call init_card
		cmp al,1
		je init_err
		mov dx,offset mmc_initokmsg
        		mov ah,9
        		int 21h
		mov dx,offset mmcwritemsg
        		mov ah,9
        		int 21h
		xor ebx,ebx
		mov bx,[ds:sector]			;there's a 32MB limit with 65536 sectors
		call write_sector_to_mmc
		cmp al,5
		je crc_err
		cmp al,6
		je write_err
		cmp al,7
		je cmd_err
		cmp al,8
		je timeout_err
		mov dx,offset mmc_ok
        		mov ah,9
        		int 21h
		jmp quit

;---------------------------------------------------------------------------------------------------------

		
read_mmc_sector:	add si,3
		call txt_to_hex
		mov [ds:sector],bx
		add si,5
		call get_filename
        		mov dx,[ds:sector]
        		mov di,offset mmcreadmsg+29
        		call hextochars
        		mov dx,offset mmc_initmsg
        		mov ah,9
        		int 21h
        		call init_card
		cmp al,1
		je init_err
		mov dx,offset mmc_initokmsg
        		mov ah,9
        		int 21h
		mov dx,offset mmcreadmsg
        		mov ah,9
        		int 21h
		xor ebx,ebx
		mov bx,[ds:sector]			;32MB limit with 65536 sectors
		call read_sector_from_mmc
		cmp al,4
		je read_err
		cmp al,9
		je dt_timeout_err
		mov dx,offset mmc_ok
        		mov ah,9
        		int 21h
		mov [word ds:save_length],512
		mov [byte ds:save_data],1
		jmp quit

;---------------------------------------------------------------------------------------------------------


read_mmc_info:	add si,3
		call get_filename
		mov dx,offset mmc_initmsg
        		mov ah,9
        		int 21h
		call init_card
		cmp al,1
		je init_err
		mov dx,offset mmc_initokmsg
        		mov ah,9
        		int 21h
		mov dx,offset mmcreadinfomsg
        		mov ah,9
        		int 21h
		call read_mmc_card_info
		cmp al,2
		je csd_err
		cmp al,3
		je cid_err
		cmp al,9
		je dt_timeout_err
		mov dx,offset mmc_ok
        		mov ah,9
        		int 21h
		mov [word ds:save_length],32		
		mov [byte ds:save_data],1

;---------------------------------------------------------------------------

quit: 		call card_power_off
				
		cmp [byte ds:save_data],1
		jne no_save
		call save_data_to_disk
		
no_save:		call save_debug_file

		mov dx,offset message2
        		mov ah,9
        		int 21h
		mov ax,4c00h           	;return to dos
        		int 21h
        
;---------------------------------------------------------------------------

init_err:		mov dx,offset mmc_error1
		jmp do_msg

csd_err:		mov dx,offset mmc_error2
		jmp do_msg
		
cid_err:		mov dx,offset mmc_error3
		jmp do_msg

read_err:		mov dx,offset mmc_error4
		jmp do_msg

crc_err:		mov dx,offset mmc_error6
		jmp do_msg

write_err:	mov dx,offset mmc_error5
		jmp do_msg
				
cmd_err:		mov dx,offset mmc_error7
		jmp do_msg

timeout_err:	mov dx,offset mmc_error8
		jmp do_msg

dt_timeout_err:	mov dx,offset mmc_error9
do_msg:  		mov ah,9
        		int 21h	
		jmp quit
				
;============================================================================================
;============================================================================================

PROC init_card	

;returns al = 0x00 if OK, 0x01 if init failed

		call card_power_off
		call wait_half_a_second
		call card_power_on
		call wait_three_milliseconds
		
		cli
		mov al,[ds:port_shadow]
		bts ax,mmc_din
		bts ax,mmc_cs
		call update_port
		call wait_half_cycle
		
		mov cx,80			;send 80 clock pulses
eighty_clocks:	bts ax,mmc_clock
		call update_port
		call wait_half_cycle
		btr ax,mmc_clock
		call update_port
		call wait_half_cycle
		loop eighty_clocks
		sti
		call select_card
			
		mov al,040h		;code for CMD 0
		mov ah,095h		;hard coded CRC (only valid for CMD 0)
		mov ebx,0			;no args
		call mmc_send_command	;send CMD 0
		call mmc_read_byte		;skip NCR
		call mmc_read_byte		;hopefully rec code 0x01 = "in idle state = 1"
		cmp al,01h
		jnz init_failed
		
		mov cx,1000		;max loops allowed to find init ready
wait_initdone:	mov al,041h		;code for CMD 1
		mov ah,000h		;null CRC
		mov ebx,0			;no args
		call mmc_send_command	;send CMD 1
		call mmc_read_byte		;skip NCR
		call mmc_read_byte
		cmp al,00h		
		jz init_done		;wait for 0x00 = "in idle state = 0"
		loop wait_initdone		;can take several hundred milliseconds

init_failed:	call deselect_card
		mov al,1
		ret

init_done:	xor al,al
		ret



ENDP 

;-------------------------------------------------------------------------------

PROC read_mmc_card_info

; returns al = 0x00 if OK, 0x02 if CSD failed, 0x03 CID if failed

		mov al,049h		;command: "read CSD" ($49)
		mov ah,000h		;dummy CRC
		mov ebx,0
		call mmc_send_command
		call mmc_read_byte		;skip NCR
		call mmc_read_byte		;get command response
		cmp al,000h
		jne end_csd_err
		
		mov cx,1000
waitdatatoken1:	call mmc_read_byte		; wait for data token $fe
		cmp al,0feh		; data packet start
		je got_data_token1
		loop waitdatatoken1	
		call deselect_card
		mov al,9
		ret		
		
got_data_token1:	mov si,offset sector_data	;get 16 bytes of CSD data
		mov cx,16
csd_loop:		call mmc_read_byte
		mov [ds:si],al
		inc si
		loop csd_loop 
		call mmc_read_byte		;CRC byte 1
		call mmc_read_byte		;CRC byte 2


		mov al,04ah		;command: "read CID" ($4a)
		mov ah,000h		;dummy CRC
		mov ebx,0
		call mmc_send_command
		call mmc_read_byte		;skip NCR
		call mmc_read_byte		;get command response
		cmp al,000h
		jne  end_cid_err

		mov cx,1000
waitdatatoken2:	call mmc_read_byte		; wait for data token $fe
		cmp al,0feh		; data packet start
		je got_data_token2
		loop waitdatatoken2	
		call deselect_card
		mov al,9
		ret		

got_data_token2:	mov si,offset sector_data+16
		mov cx,16
cid_loop:		call mmc_read_byte
		mov [ds:si],al
		inc si
		loop cid_loop 
		call mmc_read_byte		;CRC byte 1
		call mmc_read_byte		;CRC byte 2
		
		call deselect_card
		xor al,al
		ret

end_csd_err:	call deselect_card
		mov al,2
		ret
end_cid_err:	call deselect_card
		mov al,3
		ret

ENDP

;-------------------------------------------------------------------------------

PROC read_sector_from_mmc
	
; set ebx to sector location before calling	
; returns al = 00h if OK, 04h = error
	
		mov al,051h		;command: "read sector"
		mov ah,000h
		shl ebx,9			;convert sector LBA to byte address
		call mmc_send_command
		call mmc_read_byte		;skip NCR
		call mmc_read_byte		;get command response
		cmp al,000h
		jne mrs_err
		
		mov cx,1000
waitdatatoken3:	call mmc_read_byte		; wait for data token $fe
		cmp al,0feh		; data packet start
		je got_data_token3
		loop waitdatatoken3	
		mov al,9
		ret		
		
got_data_token3:	mov si,offset sector_data
		mov cx,512
mmc_rsloop:	call mmc_read_byte		;read 512 bytes
		mov [ds:si],al
		inc si
		loop mmc_rsloop 
		call mmc_read_byte		;CRC byte 1
		call mmc_read_byte		;CRC byte 2
		
		call deselect_card
		xor al,al
		ret
mrs_err:		call deselect_card
		mov al,4
		ret
ENDP

;-------------------------------------------------------------------------------

PROC write_sector_to_mmc

; set ebx to sector location before calling	
; returns al = 0 if OK, 05h= CRC error, 06h = write error, 07h = command failed


		mov al,058h
		mov ah,000h
		shl ebx,9			;convert sector LBA to byte address
		call mmc_send_command
		call mmc_read_byte		;skip NCR
		call mmc_read_byte		;get command response byte
		cmp al,000h
		jnz mmc_wscmd_failed
		mov al,0ffh
		call mmc_write_byte		;wait 8 clocks before proceding
		mov al,0feh
		call mmc_write_byte		;send data packet header code = 0xfe
		mov si,offset sector_data
		mov cx,512
mmc_wsloop:	mov al,[ds:si]
		call mmc_write_byte		;send 512 sector bytes
		inc si
		loop mmc_wsloop 
		mov al,0ffh
		call mmc_write_byte		;send null CRC byte 1
		xor al,0ffh
		call mmc_write_byte		;send null CRC byte 2
		call mmc_read_byte		;get data response
		and al,01fh
		shr al,1
		cmp al,02h
		jz mmc_wsok
		cmp ah,05h
		jz mmc_wscrce
		call deselect_card
		mov al,6			;data response = failed, write rror
		ret
mmc_wscrce:	call deselect_card
		mov al,5			;data response = failed, CRC error
		ret
mmc_wscmd_failed:	call deselect_card
		mov al,7			;command response = failed
		ret		

mmc_wsok:		mov cx,1000		
mmc_busywait:	call mmc_read_byte		;data reponse = data accepted OK
		cmp al,0ffh		;wait for busy flag to clear
		jz mmc_notbusy		;(0xff = not busy)
		loop mmc_busywait
		call deselect_card
		mov al,8			;error = time out
		ret
		
mmc_notbusy:	call deselect_card
		xor al,al			
		ret

ENDP

;-------------------------------------------------------------------------------

PROC mmc_write_byte

;set al = byte to send to mmc before calling
		
		cli			;disable interrupts
		pusha
		mov bl,al
		mov cx,8
wmb_loop:		mov al,[ds:port_shadow]
		btr ax,mmc_din
		shl bl,1			;shift out bit 7 to carry flag
		jnc wbiz
		bts ax,mmc_din		;if it was a 1, set data bit
wbiz:		call update_port		;data presented to Din
		call wait_half_cycle	;wait pulse space
		bts ax,mmc_clock		;        __
		call update_port		; ______!   clock goes high
		call wait_half_cycle	;wait pulse mark
		btr ax,mmc_clock		;__
		call update_port		;  !______ clock goes low
		loop wmb_loop		;
		bts ax,mmc_din		;return Din to normally high at end of write
		call update_port		;
		popa
		sti			;reenable interrupts
		ret
ENDP
		
;-------------------------------------------------------------------------------

PROC mmc_read_byte

;on return al = byte read from mmc

		cli			;disable interrupts
		push bx
		push cx
		push dx
		xor bl,bl

		mov cx,8
rmb_loop:		call wait_half_cycle	;wait pulse space
		shl bl,1
		mov al,[ds:port_shadow]	;   ______
		bts ax,mmc_clock		;__!
		call update_port		;clock goes high
		mov dx,pport_address+1
		in al,dx			;read in data from MMC
		bt ax,mmc_dout
		jc rmbnobit		;if input = 1, mmc data bit = 0
		or bl,01h			;if input = 0, mmc data bit = 1
rmbnobit:		call wait_half_cycle	;wait pulse mark
		mov al,[ds:port_shadow]	;__
		btr ax,mmc_clock		;  !______clock goes low
		call update_port		
		loop rmb_loop
		mov al,bl
		
		mov bx,[ds:debug_index]	;write value to debug buffer
		mov [ds:debug_buffer+bx],al
		inc bx
		mov [ds:debug_index],bx
		
		pop dx
		pop cx
		pop bx
		sti			;reenable interrupts
		ret
ENDP
		
;-------------------------------------------------------------------------------

PROC mmc_send_command

		call mmc_write_byte		;send command code byte in al
		mov ecx,ebx
		shr ecx,16
		mov al,ch
		call mmc_write_byte		;send msb of hi word
		mov al,cl
		call mmc_write_byte		;send lsb of hi word
		mov al,bh
		call mmc_write_byte		;send msb of lo word
		mov al,bl
		call mmc_write_byte		;send lsb of lo word
		mov al,ah
		call mmc_write_byte		;send CRC byte in ah
		ret
ENDP

;-------------------------------------------------------------------------------

PROC select_card

		mov al,[ds:port_shadow]
		btr ax,mmc_cs
		call update_port		;take card's CS line low
		ret
ENDP



PROC deselect_card

		mov al,[ds:port_shadow]
		bts ax,mmc_cs
		call update_port		;set CS high
		call wait_half_cycle
		mov cx,8			;send 8 clock pulses to make MMC
eight_clocks:	bts ax,mmc_clock		;release Dout
		call update_port
		call wait_half_cycle
		btr ax,mmc_clock
		call update_port
		call wait_half_cycle
		loop eight_clocks
		ret
ENDP



PROC card_power_off

		mov al,[ds:port_shadow]
		btr ax,mmc_power
		call update_port			;turn card power off
		ret
ENDP



PROC card_power_on

		mov al,[ds:port_shadow]	
		bts ax,mmc_power
		call update_port			;turn card power on
		ret
ENDP

;--------------------------------------------------------------------------------------------------------

PROC update_port
		mov dx,pport_address
		out dx,al
		mov [ds:port_shadow],al
		ret

ENDP

;============================================================================================
;============================================================================================


PROC calibrate_timer

		mov ax,3508h
		int 21h			; Get int 8 vector
		mov [ds:Old8Ofs],bx		; Store offset
		mov [ds:Old8Seg],es		; Store segment
		push ds
		mov dx,OFFSET New8IRQ	; Point to new handler
		mov ax,SEG New8IRQ
		mov ds,ax
		mov ax,2508h
		int 21h			; Set vector
		pop ds
		
		cli	
		mov al,00110100b		; Channel 0, mode 2
		out 43h,al		; Set mode
		xor ax,ax			; Zero
		jmp SHORT $+2		; Delay
		out 40h,al		; Loword 
		jmp SHORT $+2		; Delay
		out 40h,al		; Hiword 
				
		in al,21h
		mov [ds:orig_imr],al
 		mov al,0feh		;disable all but timer IRQ
       		out 21h,al  
        		sti

wait_timer:	cmp [byte ds:irq_occured],1
		jne wait_timer
		mov [byte ds:irq_occured],0
		
		xor edx,edx
wait_irq_loop:	inc edx
		cmp [ds:irq_occured],1
		jne wait_irq_loop
		shr edx,16		
		inc edx			;prevent count being zero
time_ok:		mov [ds:time_constant],edx	;loops per timer clock tick 
		
		cli                     	;restore original INT 8 settings
	        	mov al,[ds:orig_imr]
	        	out 21h,al
 		mov al,00110110b		; Channel 0, mode 3
		out 43h,al		; Set mode
		xor ax,ax			; Zero
		jmp SHORT $+2		; Delay
		out 40h,al		; Loword 
		jmp SHORT $+2		; Delay
		out 40h,al		; Hiword 
		sti

		push ds
		mov dx,[ds:Old8Ofs]
		mov ax,[ds:Old8Seg] 	
		mov ds,ax
		mov ax,2508h
		int 21h			; Restore int 8 vector
		pop ds
muh:		ret
		
ENDP		
		

PROC New8irq
		push ax 	          	;interupt handler - for timer 0
          	mov [byte ds:irq_occured],1
      		mov al,020h     		;acknowledge interrupt
        		out 020h,al
        		pop ax 
        		iret
ENDP



;--------------------------------------------------------------------------------------------------------

PROC wait_half_cycle
		
		push edx				;irqs should be disabled for stable
		push cx				;timing during this wait
		mov cx,12				;wait 12 timer clocks 
busy_loop2:	mov edx,[ds:time_constant]
busy_loop:	cmp [ds:irq_occured],1
		dec edx
		jne busy_loop
		loop busy_loop2
		pop cx
		pop edx
		ret
ENDP



PROC wait_half_a_second


		push edx				; doesnt really matter about IRQs during 
		push ecx				; this wait
		mov ecx,091ab2h
busy_loop4:	mov edx,[ds:time_constant]
busy_loop3:	cmp [ds:irq_occured],1
		dec edx
		jne busy_loop3
		dec ecx
		jne busy_loop4
		pop ecx
		pop edx
		ret

  		
ENDP


PROC wait_three_milliseconds


		push edx				; doesnt really matter about IRQs during 
		push ecx				; this wait
		mov ecx,0dfah
busy_loop6:	mov edx,[ds:time_constant]
busy_loop5:	cmp [ds:irq_occured],1
		dec edx
		jne busy_loop5
		dec ecx
		jne busy_loop6
		pop ecx
		pop edx
		ret

  		
ENDP


;-------------------------------------------------------------------------------------------------------

PROC txt_to_hex	

		mov al,[es:si]
		call conv_digit
		mov bh,al
		shl bh,4
		mov al,[es:si+1]
		call conv_digit
		or bh,al
		mov al,[es:si+2]
		call conv_digit
		mov bl,al
		shl bl,4
		mov al,[es:si+3]
		call conv_digit
		or bl,al
		ret
ENDP	
	
	
PROC conv_digit	
		sub al,030h
		cmp al,9h
		jbe gotdigit
		cmp al,016h
		ja lowcase
		sub al,07h
		jmp gotdigit
lowcase:		sub al,027h
gotdigit:		ret
		
ENDP


PROC get_filename
		mov di,offset filename
fncpy: 		mov al,[es:si]
        		cmp al," "
        		jbe fndone
        		mov [ds:di],al
        		inc di
        		inc si
        		jmp fncpy
fndone:		mov [byte ds:di],0
		ret

ENDP

;--------------------------------------------------------------------------

PROC load_data_from_disk

        		pusha
       		mov ax,03d00h
        		mov dx,offset filename
        		int 21h
        		jc lderr                	;jumps if cant find that file

        		mov bx,ds
        		push bx
        		mov [handle],ax
        		mov bx,ax
        		mov ah,3fh
        		xor al,al
        		mov cx,512
        		mov dx,offset sector_data
        		int 21h                         ;load 
        		pop bx
        		mov ds,bx
        		mov bx,[handle]
        		mov ah,3eh
        		int 21h                         ;close file
        		mov dx,offset loadokmsg
        		mov ah,9
        		int 021h
        		popa
        		xor al,al
        		ret
lderr:  		mov dx,offset loaderrmsg
        		mov ah,9
        		int 021h
        		popa
        		mov al,1
        		ret

ENDP



PROC save_data_to_disk

        		mov ah,3ch  	           ;create file
       		mov cl,00
        		mov dx,offset filename
        		int 21h
        		jnc fcrok
        		jmp fcrerr
fcrok:  		mov [ds:handle],ax
		
        		mov bx,[ds:handle]
        		mov cx,[ds:save_length]        ;length
        		push ds
        		mov dx,offset sector_data      ;start of file
        		mov ah,40h             	 ;write file
        		int 21h
        		pop ds        
        		cmp ax,cx
        		je saveok
fcrerr: 		mov dx,offset saveerrmsg
        		mov ah,9
        		int 21h
        		jmp sdone

saveok: 		mov dx,offset saveokmsg
        		mov ah,9
        		int 021h
sdone:   		ret

ENDP


PROC hextochars
	                            ;set dx to word
	        xor ax,ax           ;set di to dest for ascii
	        mov al,dh
	        shr al,4
	        mov bx,offset hexchars
	        add bx,ax
	        mov al,[ds:bx]
	        mov [ds:di],al
	        xor ax,ax
	        mov al,dh
	        and al,15
	        mov bx,offset hexchars
	        add bx,ax
	        mov al,[ds:bx]
	        mov [ds:di+1],al
	        xor ax,ax
	        mov al,dl
	        shr al,4
	        mov bx,offset hexchars
	        add bx,ax
	        mov al,[ds:bx]
	        mov [ds:di+2],al
	        xor ax,ax
	        mov al,dl
	        and al,15
	        mov bx,offset hexchars
	        add bx,ax
	        mov al,[ds:bx]
	        mov [ds:di+3],al
	        ret

ENDP hextochars  



PROC save_debug_file

		cmp [word ds:debug_index],0
		jne savedb
		ret

savedb:     	mov ah,3ch  	           ;create file
       		mov cl,00
        		mov dx,offset debug_filename
        		int 21h
        		jnc fcrok2
        		jmp fcrerr2
fcrok2:  		mov [ds:handle],ax
		
        		mov bx,[ds:handle]
        		mov cx,[ds:debug_index]        ;length
        		push ds
        		mov dx,offset debug_buffer     ;start of file
        		mov ah,40h             	 ;write file
        		int 21h
        		pop ds        
        		cmp ax,cx
        		je saveok2
fcrerr2: 		mov dx,offset saveerrmsg
        		mov ah,9
        		int 21h
        		jmp sdone2

saveok2: 		mov dx,offset dbsaveokmsg
        		mov ah,9
        		int 021h
sdone2:  		ret

ENDP

   		
;---------------------------------------------------------------------------
END start
;---------------------------------------------------------------------------

